﻿using System;
using System.Collections.Generic;
using GameUtils.Graphics.Data;
using GameUtils.Math;

namespace GameUtils.Graphics
{
    public class Geometry : Resource<GeometryData>
    {
        readonly List<GeometryOperation> operations;
        bool changed;
        int changeIndex;
        readonly object locker;

        /// <summary>
        /// Adds an arc to the current figure.
        /// </summary>
        public void AddArc(Arc arc)
        {
            lock (locker)
            {
                operations.Add(new AddArcOperation(arc));
                changed = true;
            }
        }

        /// <summary>
        /// Adds a line to the current figure.
        /// </summary>
        public void AddLine(Vector2 point)
        {
            lock (locker)
            {
                operations.Add(new AddLineOperation(point));
                changed = true;
            }
        }

        /// <summary>
        /// Adds multiple lines to the current figure.
        /// </summary>
        public void AddLines(Vector2[] points)
        {
            if (points == null)
                throw new ArgumentNullException("points");
            if (points.Length == 0)
                throw new ArgumentException("There has to be at least one point in the list.", "points");

            lock (locker)
            {
                operations.Add(new AddLinesOperation(points));
                changed = true;
            }
        }

        /// <summary>
        /// Starts a new figure.
        /// </summary>
        /// <remarks>This method has to be called before any other figure-dependend method gets called.</remarks>
        public void BeginFigure(Vector2 point, FigureBegin begin)
        {
            lock (locker)
            {
                operations.Add(new BeginFigureOperation(point, begin));
                changed = true;
            }
        }

        /// <summary>
        /// Closes the currently open figure.
        /// </summary>
        /// <remarks>This method has to be called before the geometry can be closed.</remarks>
        public void EndFigure(FigureEnd end)
        {
            lock (locker)
            {
                operations.Add(new EndFigureOperation(end));
                changed = true;
            }
        }

        /// <summary>
        /// Adds a bezier curve to the current figure.
        /// </summary>
        public void AddBezier(Bezier bezier)
        {
            lock (locker)
            {
                operations.Add(new AddBezierOperation(bezier));
                changed = true;
            }
        }

        /// <summary>
        /// Adds multiple bezier curves to the current figure.
        /// </summary>
        public void AddBeziers(Bezier[] beziers)
        {
            if (beziers == null)
                throw new ArgumentNullException("beziers");
            if (beziers.Length == 0)
                throw new ArgumentException("There has to be at least one element in the list.", "beziers");

            lock (locker)
            {
                operations.Add(new AddBeziersOperation(beziers));
                changed = true;
            }
        }

        /// <summary>
        /// Closes this geometry.
        /// </summary>
        /// <remarks>After a geometry has been closed it is only guaranteed of being able to transform but a non-closed geometry isn't guaranteed to render.</remarks>
        public void Close()
        {
            lock (locker)
            {
                operations.Add(new CloseOperation());
                changed = true;
            }
        }

        /// <summary>
        /// Adds a cardinal spline curve to the current figure.
        /// </summary>
        public void AddCurve(Vector2[] points, float tension = 0.5f)
        {
            if (points == null)
                throw new ArgumentNullException("points");
            if (points.Length == 0)
                throw new ArgumentException("There has to be at least one point in the list.", "points");

            lock (locker)
            {
                operations.Add(new AddCurveOperation(points, tension));
                changed = true;
            }
        }

        /// <summary>
        /// Adds another geometry to this one.
        /// </summary>
        public void CombineWith(Geometry geometry, GeometryCombineMode combineMode)
        {
            if (geometry == null)
                throw new ArgumentNullException("geometry");

            lock (locker)
            {
                operations.Add(new CombineOperation(geometry, combineMode));
                changed = true;
            }
        }

        /// <summary>
        /// Adds a rectangle to the geometry.
        /// </summary>
        public void AddRectangle(Rectangle rect)
        {
            lock (locker)
            {
                operations.Add(new AddRectangleOperation(rect));
                changed = true;
            }
        }

        /// <summary>
        /// Adds an ellipse to the geometry.
        /// </summary>
        public void AddEllipse(Ellipse ellipse)
        {
            lock (locker)
            {
                operations.Add(new AddEllipseOperation(ellipse));
                changed = true;
            }
        }

        /// <summary>
        /// Transforms this geometry by a given matrix.
        /// </summary>
        public void Transform(Matrix2x3 matrix)
        {
            if (matrix.IsIdentity)
                return;

            lock (locker)
            {
                operations.Add(new TransformOperation(matrix));
                changed = true;
            }
        }

        /// <summary>
        /// Creates a new geometry.
        /// </summary>
        public Geometry()
        {
            operations = new List<GeometryOperation>();
            locker = new object();
        }

        protected override void ApplyChanges()
        {
            lock (locker)
            {
                if (changed)
                {
                    for (int i = changeIndex; i < operations.Count; i++)
                        operations[i].ApplyTo(Data);
                    changeIndex = operations.Count;
                    changed = false;
                }
            }
        }

        protected override GeometryData CreateData(Factory factory)
        {
            lock (locker)
            {
                changed = false;
                GeometryData data = factory.CreateGeometry();
                for (int i = 0; i < operations.Count; i++)
                    operations[i].ApplyTo(data);
                changeIndex = operations.Count;
                return data;
            }
        }

        private abstract class GeometryOperation
        {
            public abstract void ApplyTo(GeometryData geometry);
        }

        private class AddArcOperation : GeometryOperation
        {
            readonly Arc arc;

            public AddArcOperation(Arc arc)
            {
                this.arc = arc;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddArc(arc);
            }
        }

        private class AddLineOperation : GeometryOperation
        {
            readonly Vector2 point;

            public AddLineOperation(Vector2 point)
            {
                this.point = point;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddLine(point);
            }
        }

        private class AddLinesOperation : GeometryOperation
        {
            readonly Vector2[] points;

            public AddLinesOperation(Vector2[] points)
            {
                this.points = points;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddLines(points);
            }
        }

        private class BeginFigureOperation : GeometryOperation
        {
            readonly Vector2 point;
            readonly FigureBegin begin;

            public BeginFigureOperation(Vector2 point, FigureBegin begin)
            {
                this.point = point;
                this.begin = begin;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.BeginFigure(point, begin);
            }
        }

        private class EndFigureOperation : GeometryOperation
        {
            readonly FigureEnd end;

            public EndFigureOperation(FigureEnd end)
            {
                this.end = end;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.EndFigure(end);
            }
        }

        private class AddBezierOperation : GeometryOperation
        {
            readonly Bezier bezier;

            public AddBezierOperation(Bezier bezier)
            {
                this.bezier = bezier;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddBezier(bezier);
            }
        }

        private class AddBeziersOperation : GeometryOperation
        {
            readonly Bezier[] beziers;

            public AddBeziersOperation(Bezier[] beziers)
            {
                this.beziers = beziers;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddBeziers(beziers);
            }
        }

        private class CloseOperation : GeometryOperation
        {
            public override void ApplyTo(GeometryData geometry)
            {
                geometry.Close();
            }
        }

        private class AddCurveOperation : GeometryOperation
        {
            readonly Vector2[] points;
            readonly float tension;

            public AddCurveOperation(Vector2[] points, float tension)
            {
                this.points = points;
                this.tension = tension;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddCurve(points, tension);
            }
        }

        private class CombineOperation : GeometryOperation
        {
            readonly Geometry geometry;
            readonly GeometryCombineMode combineMode;

            public CombineOperation(Geometry geometry, GeometryCombineMode combineMode)
            {
                this.geometry = geometry;
                this.combineMode = combineMode;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddGeometry(this.geometry, combineMode);
            }
        }

        private class AddRectangleOperation : GeometryOperation
        {
            readonly Rectangle rect;

            public AddRectangleOperation(Rectangle rect)
            {
                this.rect = rect;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddRectangle(rect);
            }
        }

        private class AddEllipseOperation : GeometryOperation
        {
            readonly Ellipse ellipse;

            public AddEllipseOperation(Ellipse ellipse)
            {
                this.ellipse = ellipse;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.AddEllipse(ellipse);
            }
        }

        private class TransformOperation : GeometryOperation
        {
            readonly Matrix2x3 matrix;

            public TransformOperation(Matrix2x3 matrix)
            {
                this.matrix = matrix;
            }

            public override void ApplyTo(GeometryData geometry)
            {
                geometry.Transform(matrix);
            }
        }
    }
}
